/*	$NetBSD$	*/
/*-
 * Copyright (c) 2009 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by Joerg Sonnenberger.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * See http://burtleburtle.net/bob/hash/doobs.html for the full description
 * and the original version of the code.  This version differs by exposing
 * the full internal state and avoiding byte operations in the inner loop
 * if the key is aligned correctly.
 */

#include <sys/endian.h>

#include "jenkins_hash.h"

#define mix(a, b, c) do {		\
	a -= b; a -= c; a ^= (c >> 13);	\
	b -= c; b -= a; b ^= (a << 8);	\
	c -= a; c -= b; c ^= (b >> 13);	\
	a -= b; a -= c; a ^= (c >> 12);	\
	b -= c; b -= a; b ^= (a << 16);	\
	c -= a; c -= b; c ^= (b >> 5);	\
	a -= b; a -= c; a ^= (c >> 3);	\
	b -= c; b -= a; b ^= (a << 10);	\
	c -= a; c -= b; c ^= (b >> 15);	\
} while (/* CONSTCOND */0)

#define FIXED_SEED	0x9e3779b9	/* Golden ratio, arbitrary constant */

void
jenkins_vector_hash(const void * __restrict key, size_t len, uint32_t seed,
    uint32_t hashes[3])
{
	uint32_t orig_len;
	const uint8_t *k;

	orig_len = (uint32_t)len;

	hashes[0] = hashes[1] = FIXED_SEED;
	hashes[2] = seed;

	if ((uintptr_t)key & 3) {
		k = key;
		while (len >= 12) {
			hashes[0] += le32dec(k);
			hashes[1] += le32dec(k + 4);
			hashes[2] += le32dec(k + 8);
			mix(hashes[0], hashes[1], hashes[2]);
			k += 12;
			len -= 12;
		}
	} else {
		const uint32_t *key32 = key;
		while (len >= 12) {
			hashes[0] += le32toh(key32[0]);
			hashes[1] += le32toh(key32[1]);
			hashes[2] += le32toh(key32[2]);
			mix(hashes[0], hashes[1], hashes[2]);
			key32 += 3;
			len -= 12;
		}
		k = (const uint8_t *)key32;
	}
	hashes[2] += orig_len;

	switch (len) {
	case 11:
		hashes[2] += (uint32_t)k[10] << 24;
		/* FALLTHROUGH */
	case 10:
		hashes[2] += (uint32_t)k[9] << 16;
		/* FALLTHROUGH */
	case 9:
		hashes[2] += (uint32_t)k[8] << 8;
		/* FALLTHROUGH */
	case 8:
		hashes[1] += (uint32_t)k[7] << 24;
		/* FALLTHROUGH */
	case 7:
		hashes[1] += (uint32_t)k[6] << 16;
		/* FALLTHROUGH */
	case 6:
		hashes[1] += (uint32_t)k[5] << 8;
		/* FALLTHROUGH */
	case 5:
		hashes[1] += k[4];
		/* FALLTHROUGH */
	case 4:
		hashes[0] += (uint32_t)k[3] << 24;
		/* FALLTHROUGH */
	case 3:
		hashes[0] += (uint32_t)k[2] << 16;
		/* FALLTHROUGH */
	case 2:
		hashes[0] += (uint32_t)k[1] << 8;
		/* FALLTHROUGH */
	case 1:
		hashes[0] += k[0];
		/* FALLTHROUGH */
	}
	mix(hashes[0], hashes[1], hashes[2]);
}

uint32_t
jenkins_hash(const void * __restrict key, size_t len, uint32_t seed)
{
	uint32_t hashes[3];

	jenkins_vector_hash(key, len, seed, hashes);
	return hashes[0];
}

