/*	$NetBSD$	*/
/*-
 * Copyright (c) 2009 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by Joerg Sonnenberger.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__RCSID("$NetBSD$");

#include <err.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "jenkins_hash.h"

/*
 * A full description of the algorithm can be found in:
 * "An optimal algorithm for generating minimal perfect hash functions"
 * by Czech, Havas and Majewski in Information Processing Letters,
 * 43(5):256-264, October 1992.
 */

/*
 * The algorithm is based on random, acyclic graphs.
 *
 * Each edge in the represents a key.  The vertices are the reminder of
 * the hash function mod n.  n = cm with c > 2, otherwise the propability
 * of finding an acyclic graph is very low.
 *
 * After the hashing phase, the graph is checked for cycles.
 * A cycle-free graph is either empty or has a vertex of degree 1.
 * Removing this vertex doesn't change this property, so applying this
 * recursively reduces the size of the graph.  If the graph is empty at
 * the end of the process, it was acyclic.
 *
 * The assignment step now sets g[i] := 0 for one vertex of each
 * connected component.  traverse follows the pathes from this vertex
 * and the visited check makes sure that g[i] is set exactly once.
 */

static const uint32_t unused = 0xffffffffU;
static FILE *output;

struct vertex {
	uint32_t l_edge, r_edge;
};

struct edge {
	uint32_t left, right;
	uint32_t l_prev, l_next;
	uint32_t r_prev, r_next;
};

struct state {
	struct vertex *verts;
	struct edge *edges;
	uint32_t *g;
	uint8_t *visited;
	uint32_t seed, m, n;
};

static int
init_graph(struct state *state, const void * __restrict * keys,
    const size_t * __restrict keylens)
{
	size_t i;

	for (i = 0; i < state->m; ++i) {
		uint32_t hashes[3];

		jenkins_vector_hash(keys[i], keylens[i], state->seed, hashes);
		state->edges[i].left = hashes[0] % state->n;
		state->edges[i].right = hashes[1] % state->n;
		if (state->edges[i].left == state->edges[i].right)
			return -1;
	}
	return 0;
}

static void
build_graph(struct state *state)
{
	struct vertex *v;
	size_t i;

	for (i = 0; i < state->n; ++i) {
		state->verts[i].l_edge = unused;
		state->verts[i].r_edge = unused;
	}

	for (i = 0; i < state->m; ++i) {
		v = &state->verts[state->edges[i].left];
		if (v->l_edge != unused)
			state->edges[v->l_edge].l_prev = i;
		state->edges[i].l_next = v->l_edge;
		state->edges[i].l_prev = unused;
		v->l_edge = i;
		v = &state->verts[state->edges[i].right];
		if (v->r_edge != unused)
			state->edges[v->r_edge].r_prev = i;
		state->edges[i].r_next = v->r_edge;
		state->edges[i].r_prev = unused;
		v->r_edge = i;
	}
}

static void
remove_vertex(struct state *state, struct vertex *v)
{
	struct edge *e;
	struct vertex *v2;

	for (;;) {
		if (v->l_edge != unused && v->r_edge != unused)
			break;
		if (v->l_edge == unused && v->r_edge == unused)
			break;

		if (v->l_edge != unused) {
			e = &state->edges[v->l_edge];
			if (e->l_next != unused)
				break;
			v->l_edge = unused; /* No other elements possible! */
			v2 = &state->verts[e->right];
			if (e->r_prev == unused)
				v2->r_edge = e->r_next;
			else
				state->edges[e->r_prev].r_next = e->r_next;
			if (e->r_next != unused)
				state->edges[e->r_next].r_prev = e->r_prev;
			v = v2;
		} else {
			e = &state->edges[v->r_edge];
			if (e->r_next != unused)
				break;
			v->r_edge = unused; /* No other elements possible! */
			v2 = &state->verts[e->left];
			if (e->l_prev == unused)
				v2->l_edge = e->l_next;
			else
				state->edges[e->l_prev].l_next = e->l_next;
			if (e->l_next != unused)
				state->edges[e->l_next].l_prev = e->l_prev;
			v = v2;
		}
	}
}

static int
check_acyclic(struct state *state)
{
	size_t i;

	for (i = 0; i < state->n; ++i)
		remove_vertex(state, &state->verts[i]);

	for (i = 0; i < state->n; ++i) {
		if (state->verts[i].l_edge != unused ||
		    state->verts[i].r_edge != unused)
			return -1;
	}
	return 0;
}

static void
traverse(struct state *state, uint32_t v)
{
	struct edge *e;
	uint32_t e_idx;
	uint32_t v2;

	state->visited[v] = 1;
	for (e_idx = state->verts[v].l_edge; e_idx != unused;
	     e_idx = e->l_next) {
		e = &state->edges[e_idx];
		v2 = e->right;
		if (state->visited[v2])
			continue;
		state->g[v2] = (state->m + e_idx - state->g[v]) % state->m;
		traverse(state, v2);			
	}
	for (e_idx = state->verts[v].r_edge; e_idx != unused;
	     e_idx = e->r_next) {
		e = &state->edges[e_idx];
		v2 = e->left;
		if (state->visited[v2])
			continue;
		state->g[v2] = (state->m + e_idx - state->g[v]) % state->m;
		traverse(state, v2);
	}
}

static void
assign_nodes(struct state *state)
{
	size_t i;

	for (i = 0; i < state->n; ++i) {
		if (state->visited[i])
			continue;
		state->g[i] = 0;
		traverse(state, i);
	}
}

static void
print_hash(struct state *state)
{
	size_t i;
	const char *g_type;

	fprintf(output, "uint32_t\n");
	fprintf(output, "hash(const void * __restrict key, size_t keylen)\n");
	fprintf(output, "{\n");
	if (state->n >= 65536)
		g_type = "uint32_t";
	else if (state->n >= 256)
		g_type = "uint16_t";
	else
		g_type = "uint8_t";
	fprintf(output, "\tstatic const %s g[%" PRId32 "] = {\n", g_type, state->n);
	for (i = 0; i < state->n; ++i) {
		if (i % 4 == 0)
			fprintf(output, "\t    ");
		if (i % 4 == 3)
			fprintf(output, "0x%08" PRIx32"U,\n", state->g[i]);
		else
			fprintf(output, "0x%08" PRIx32"U, ", state->g[i]);
	}
	if (i % 4 == 3)
		fprintf(output, "\t};\n");
	else
		fprintf(output, "\n\t};\n");
	fprintf(output, "\tuint32_t h[3];\n\n");
	fprintf(output, "\tjenkins_vector_hash(key, keylen, 0x%08" PRIx32 "U, h);\n",
	    state->seed);
	fprintf(output, "\treturn (g[h[0] %% %" PRIu32 "] + "
	    "g[h[1] %% %" PRIu32"]) %% %" PRIu32 ";\n",
	    state->n, state->n, state->m);
	fprintf(output, "}\n");
}

static int
chm_compute(size_t m, const void * __restrict * keys,
    const size_t * __restrict keylens, double c)
{
	struct state state;
	int retval = -1;

	state.seed = arc4random();
	state.m = m;
	state.n = c * m;
	if (state.n == 2 * m)
		++state.n;
	state.verts = calloc(sizeof(struct vertex), state.n);
	state.edges = calloc(sizeof(struct edge), state.m);
	state.g = calloc(sizeof(uint32_t), state.n);
	state.visited = calloc(sizeof(uint8_t), state.n);

	if (init_graph(&state, keys, keylens))
		goto failed;
	build_graph(&state);
	if (check_acyclic(&state))
		goto failed;
	build_graph(&state);
	assign_nodes(&state);
	print_hash(&state);

#ifdef DEBUG
	for (size_t i = 0; i < state.m; ++i)
		if (i != (state.g[state.edges[i].left] + state.g[state.edges[i].right]) % state.m)
			abort();
#endif

	retval = 0;

failed:
	free(state.verts);
	free(state.edges);
	free(state.g);
	free(state.visited);
	return retval;
}

static __dead
void usage(void)
{
	fprintf(stderr,
	    "%s [-c utilisation] [-i iterations] [-o output] input\n",
	    getprogname());
	exit(1);
}

int
main(int argc, char **argv)
{
	FILE *input;
	size_t curlen = 0, curalloc = 0;
	char *line, *eos;
	size_t line_len;
	const void **keys = NULL;
	size_t *keylens = NULL;
	uint32_t max_iterations = 0xffffffU;
	long long tmp;
	int looped, ch;
	double c = 2;

	while ((ch = getopt(argc, argv, "c:i:o:")) != -1) {
		switch (ch) {
		case 'c':
			errno = 0;
			c = strtod(optarg, &eos);
			if (errno || eos[0] || c < 2 || c > 10)
				errx(2, "The utilisation constant c must be "
				    "between 2 and 10");
			break;
		case 'i':
			errno = 0;
			tmp = strtoll(optarg, &eos, 0);
			if (errno || eos == optarg || eos[0] ||
			    tmp < 0 || tmp > 0xffffffffU)
				errx(2, "Iteration count must be "
				    "a 32bit integer");
			max_iterations = (uint32_t)tmp;
			break;
		case 'o':
			if (output)
				fclose(output);
			output = fopen(optarg, "w");
			if (output == NULL)
				err(2, "cannot open output file");
			break;
		default:
			usage();
		}
	}

	argc -= optind;
	argv += optind;

	if (argc > 1)
		usage();

	if (argc == 1) {
		input = fopen(argv[0], "r");
		if (input == NULL)
			err(1, "can't open input file");
	} else
		input = stdin;

	if (output == NULL)
		output = stdout;

	while ((line = fgetln(input, &line_len)) != NULL) {
		if (line_len && line[line_len - 1] == '\n')
			--line_len;
		if (curlen == curalloc) {
			if (curalloc < 256)
				curalloc = 256;
			else
				curalloc += curalloc;
			keys = realloc(keys, curalloc * sizeof(*keys));
			if (keys == NULL)
				err(1, "realloc failed");
			keylens = realloc(keylens,
			    curalloc * sizeof(*keylens));
			if (keylens == NULL)
				err(1, "realloc failed");
		}
		if ((keys[curlen] = strndup(line, line_len)) == NULL)
			err(1, "malloc failed");
		keylens[curlen] = line_len;
		++curlen;
	}

	if (input != stdin)
		fclose(input);

	looped = 0;
	while (chm_compute(curlen, keys, keylens, c)) {
		fputc('.', stderr);
		looped = 1;
		if (max_iterations == 0xffffffffU)
			continue;
		if (--max_iterations == 0) {
			fputc('\n', stderr);
			errx(1, "Iteration count reached");
		}
	}
	if (looped)
		fputc('\n', stderr);

	return 0;
}
